#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <utime.h>
#include "fallg.h"

struct fileinfo fi;

#ifdef __INTEL__
void turn_info (void);
#endif

main (int argc, char *argv[])
{
char *fname, *inputf, *buffer, ext[5];
int count, fd, fd1;
struct stat st;
struct utimbuf ut;

	argc--;
	argv++;

	if (!argc)
	{
	   fprintf (stderr, "Missing input file name!\n");
	   exit (1);
	}

	inputf = strdup (*argv);

	if ((fname = malloc (strlen (inputf) + 5)) == NULL)
	{
	   fprintf (stderr, "Not enough memory to hold a filename!\n");
	   free (inputf);
	   exit (1);
	}

	count = 0;

	if ((fd = open (inputf, O_RDWR | O_CREAT | O_EXCL, 0660)) == -1)
	{
	   fprintf (stderr, "Error creating file %s!\nMaybe file already exists?\n", inputf);
	   free (inputf);
	   free (fname);
	   exit (1);
	}

	while (1)
	{
	   strcpy (fname, inputf);
	   sprintf (&ext[0], ".%03d", count);
	   strcat (fname, ext);
#ifdef DEBUG
	   printf ("Versuche Datei %s zu oeffnen...\n", fname);
#endif
	   if ((fd1 = open (fname, O_RDONLY)) == -1)
	   {
	      if (!count)
	      {
		 fprintf (stderr, "Can not find first \"split\" file!\n");
		 free (fname);
		 free (inputf);
		 exit (1);
	      }
#ifdef DEBUG
	      printf ("Schleife wurde bei Datei %s abgebrochen!\n", fname);
#endif
	      break;
	   }

	   fstat (fd1, &st);

	   if ((buffer = (char *)malloc (st.st_size + 1)) == NULL)
	   {
	      fprintf (stderr, "Not enough memory for buffer!\n");
	      free (inputf);
	      free (fname);
	      exit (1);
	   }
#ifdef DEBUG
	   printf ("Buffer mit %ld laenge wurde allokiert!\n", st.st_size);
#endif
	   read (fd1, &fi, sizeof (struct fileinfo));
#ifdef __INTEL__
	   turn_info ();
#endif
	   if (fi.fi_count != count)
	   {
	      fprintf (stderr, "File %s is corrupt!\n", fname);
	      free (buffer);
	      free (fname);
	      free (inputf);
	      exit (1);
	   }

	   printf ("Merging file %s...\r", fname);
	   fflush (stdout);
	   read (fd1, buffer, st.st_size - sizeof (struct fileinfo));
	   write (fd, buffer, st.st_size - sizeof (struct fileinfo));
	   close (fd1);
	   free (buffer);
	   count++;
	}

	fchmod (fd, fi.fi_mode);
	close (fd);
	ut.actime = fi.fi_atime;
	ut.modtime = fi.fi_mtime;
	utime (inputf, &ut);
	printf ("\nMerged %d files together and build %s.\n", count, inputf);
	free (inputf);
	free (fname);
	exit (0);
}

#ifdef __INTEL__
void turn_info ()
{
unsigned char b1, b2, b3, b4;

	/* fi_mode */
#ifdef DEBUG
	printf ("Ganze Zahl: %08lx\n", fi.fi_mode);
#endif
	b1 = (unsigned char)(fi.fi_mode & 0x000000ff);
	b2 = (unsigned char)((fi.fi_mode & 0x0000ff00) >> 8);
	b3 = (unsigned char)((fi.fi_mode & 0x00ff0000) >> 16);
	b4 = (unsigned char)((fi.fi_mode & 0xff000000) >> 24);
	fi.fi_mode = (imode_t)(((imode_t)b1 << 24) | ((imode_t)b2 << 16) |
				((imode_t)b3 << 8) | (imode_t)b4);
#ifdef DEBUG
	printf ("Einzelzahlen: %02x %02x %02x %02x\n", (int)b4, (int)b3, (int)b2, (int)b1);
	printf ("Neue Zahl   : %08lx\n", fi.fi_mode);
#endif

	/* fi_uid */
	b1 = (unsigned char)(fi.fi_uid & 0x000000ff);
	b2 = (unsigned char)((fi.fi_uid & 0x0000ff00) >> 8);
	b3 = (unsigned char)((fi.fi_uid & 0x00ff0000) >> 16);
	b4 = (unsigned char)((fi.fi_uid & 0xff000000) >> 24);
	fi.fi_uid = (long)(((long)b1 << 24) | ((long)b2 << 16) |
				((long)b3 << 8) | (long)b4);

	/* fi_gid */
	b1 = (unsigned char)(fi.fi_gid & 0x000000ff);
	b2 = (unsigned char)((fi.fi_gid & 0x0000ff00) >> 8);
	b3 = (unsigned char)((fi.fi_gid & 0x00ff0000) >> 16);
	b4 = (unsigned char)((fi.fi_gid & 0xff000000) >> 24);
	fi.fi_gid = (long)(((long)b1 << 24) | ((long)b2 << 16) |
				((long)b3 << 8) | (long)b4);

	/* fi_atime */
	b1 = (unsigned char)(fi.fi_atime & 0x000000ff);
	b2 = (unsigned char)((fi.fi_atime & 0x0000ff00) >> 8);
	b3 = (unsigned char)((fi.fi_atime & 0x00ff0000) >> 16);
	b4 = (unsigned char)((fi.fi_atime & 0xff000000) >> 24);
	fi.fi_atime = (long)(((long)b1 << 24) | ((long)b2 << 16) |
				((long)b3 << 8) | (long)b4);

	/* fi_mtime */
	b1 = (unsigned char)(fi.fi_mtime & 0x000000ff);
	b2 = (unsigned char)((fi.fi_mtime & 0x0000ff00) >> 8);
	b3 = (unsigned char)((fi.fi_mtime & 0x00ff0000) >> 16);
	b4 = (unsigned char)((fi.fi_mtime & 0xff000000) >> 24);
	fi.fi_mtime = (long)(((long)b1 << 24) | ((long)b2 << 16) |
				((long)b3 << 8) | (long)b4);

	/* fi_ctime */
	b1 = (unsigned char)(fi.fi_ctime & 0x000000ff);
	b2 = (unsigned char)((fi.fi_ctime & 0x0000ff00) >> 8);
	b3 = (unsigned char)((fi.fi_ctime & 0x00ff0000) >> 16);
	b4 = (unsigned char)((fi.fi_ctime & 0xff000000) >> 24);
	fi.fi_ctime = (long)(((long)b1 << 24) | ((long)b2 << 16) |
				((long)b3 << 8) | (long)b4);

	/* fi_count */
	b1 = (unsigned char)(fi.fi_count & 0x000000ff);
	b2 = (unsigned char)((fi.fi_count & 0x0000ff00) >> 8);
	b3 = (unsigned char)((fi.fi_count & 0x00ff0000) >> 16);
	b4 = (unsigned char)((fi.fi_count & 0xff000000) >> 24);
	fi.fi_count = (long)(((long)b1 << 24) | ((long)b2 << 16) |
				((long)b3 << 8) | (long)b4);
}
#endif	/* __INTEL__ */
