#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "fallg.h"

/* #define DEBUG */
#define DEFLEN	(1433600L - (long)sizeof (struct fileinfo))

struct fileinfo fi;

long get_filelen (char *file);
#ifdef __INTEL__
void turn_info (void);
#endif

main (int argc, char *argv[])
{
long maxlen = DEFLEN;
long reallen, i, pos, j, a;
int fd, fd1, count;
char *inputf, *buffer, *fname;
char ext[5];

	argc--;
	argv++;

	if (!argc)
	{
	   fprintf (stderr, "Missing name of input file!\n");
	   exit (1);
	}

	if (strcmp ("-s", *argv) == 0)
	{
	   argc--;
	   argv++;

	   if (!argc)
	   {
	      fprintf (stderr, "Missing size of output files!\n");
	      exit (1);
	   }

	   maxlen = atol (*argv);
#ifdef DEBUG
	   printf ("Maxlen = %ld\n", maxlen);
#endif

	   if (maxlen <= ((long)sizeof (struct fileinfo) + 1))
	   {
	      fprintf (stderr, "Output file size too small!\n");
	      exit (1);
	   }

	   maxlen -= (long)sizeof (struct fileinfo);
	   argc--;
	   argv++;

	   if (!argc)
	   {
	      fprintf (stderr, "Missing name of input file!\n");
	      exit (1);
	   }
	}

	inputf = strdup (*argv);
	reallen = get_filelen (inputf);
#ifdef DEBUG
	printf ("Inputname: %s\n", inputf);
	printf ("Tatsaechliche Dateilaenge: %ld\n", reallen);
#endif

	if (reallen == -1L)
	{
	   fprintf (stderr, "Unable to access file %s!\n", inputf); 
	   free (inputf);
	   exit (1);
	}

	if (reallen <= maxlen)
	{
	   fprintf (stderr, "Size of file %s is smaller or equal to given size of %ld!\n", inputf, maxlen);
	   free (inputf);
	   exit (1);
	}

	if ((buffer = (char *)malloc (maxlen + 1L)) == NULL)
	{
	   fprintf (stderr, "Not enough memory for buffer!");
	   free (inputf);
	   exit (1);
	}

	if ((fd = open (inputf, O_RDONLY)) == -1)
	{
	   fprintf (stderr, "Error opening file %s!", inputf);
	   free (inputf);
	   free (buffer);
	   exit (1);
	}

	i = pos = 0L;
	count = 0;

	if ((fname = (char *)malloc (strlen (inputf) + 5)) == NULL)
	{
	   fprintf (stderr, "Not enough memory for holding a filename!\n");
	   free (inputf);
	   free (buffer);
	   exit (1);
	}

	while (i < reallen)
	{
	   read (fd, buffer, maxlen);
	   j = lseek (fd, 0L, SEEK_CUR);
#ifdef DEBUG
	   printf ("Stehe auf Position: %ld\n", j);
#endif
	   if (j > 0L)
	   {
	      a = j - pos;
	      pos = j;
	   }

	   sprintf (&ext[0], ".%03d", count);
	   strcpy (fname, inputf);
	   strcat (fname, ext);

	   if ((fd1 = open (fname, O_RDWR | O_CREAT | O_TRUNC, 0660)) == -1)
	   {
	      fprintf (stderr, "Error creating file %s!\n", fname);
	      free (fname);
	      free (buffer);
	      free (inputf);
	      exit (1);
	   }
#ifdef DEBUG
	   printf ("POS: %ld, I: %ld, A: %ld\n", pos, i, a);
#endif
	   printf ("Writing part %s...\r", fname);
	   fflush (stdout);

	   /*
	    * First we write a info header to the file
	    */
	   fi.fi_count = count;
#ifdef __INTEL__
	   turn_info ();
#endif
	   write (fd1, &fi, sizeof (struct fileinfo));
#ifdef __INTEL__
	   /*
	    * We have to turn it back to make it work for the
	    * next time.
	    */
	   turn_info ();
#endif
	   write (fd1, buffer, a);

	   close (fd1);
	   count++;
	   i += maxlen;
	}

	printf ("\nFile %s splitted into %d parts.\n", inputf, count);
	close (fd);
	free (inputf);
	free (buffer);
	free (fname);
	exit (0);
}

long get_filelen (char *file)
{
struct stat st;

	if (stat (file, &st) == -1)
	   return -1L;

	fi.fi_mode = st.st_mode;
	fi.fi_uid = st.st_uid;
	fi.fi_gid = st.st_gid;
	fi.fi_atime = st.st_atime;
	fi.fi_mtime = st.st_mtime;
	fi.fi_ctime = st.st_ctime;

	return st.st_size;
}

#ifdef __INTEL__
void turn_info ()
{
unsigned char b1, b2, b3, b4;

	/* fi_mode */
#ifdef DEBUG
	printf ("Ganze Zahl: %08lx\n", fi.fi_mode);
#endif
	b1 = (unsigned char)(fi.fi_mode & 0x000000ff);
	b2 = (unsigned char)((fi.fi_mode & 0x0000ff00) >> 8);
	b3 = (unsigned char)((fi.fi_mode & 0x00ff0000) >> 16);
	b4 = (unsigned char)((fi.fi_mode & 0xff000000) >> 24);
	fi.fi_mode = (imode_t)(((imode_t)b1 << 24) | ((imode_t)b2 << 16) |
				((imode_t)b3 << 8) | (imode_t)b4);
#ifdef DEBUG
	printf ("Einzelzahlen: %02x %02x %02x %02x\n", (int)b4, (int)b3, (int)b2, (int)b1);
	printf ("Neue Zahl   : %08lx\n", fi.fi_mode);
#endif

	/* fi_uid */
	b1 = (unsigned char)(fi.fi_uid & 0x000000ff);
	b2 = (unsigned char)((fi.fi_uid & 0x0000ff00) >> 8);
	b3 = (unsigned char)((fi.fi_uid & 0x00ff0000) >> 16);
	b4 = (unsigned char)((fi.fi_uid & 0xff000000) >> 24);
	fi.fi_uid = (long)(((long)b1 << 24) | ((long)b2 << 16) |
				((long)b3 << 8) | (long)b4);

	/* fi_gid */
	b1 = (unsigned char)(fi.fi_gid & 0x000000ff);
	b2 = (unsigned char)((fi.fi_gid & 0x0000ff00) >> 8);
	b3 = (unsigned char)((fi.fi_gid & 0x00ff0000) >> 16);
	b4 = (unsigned char)((fi.fi_gid & 0xff000000) >> 24);
	fi.fi_gid = (long)(((long)b1 << 24) | ((long)b2 << 16) |
				((long)b3 << 8) | (long)b4);

	/* fi_atime */
	b1 = (unsigned char)(fi.fi_atime & 0x000000ff);
	b2 = (unsigned char)((fi.fi_atime & 0x0000ff00) >> 8);
	b3 = (unsigned char)((fi.fi_atime & 0x00ff0000) >> 16);
	b4 = (unsigned char)((fi.fi_atime & 0xff000000) >> 24);
	fi.fi_atime = (long)(((long)b1 << 24) | ((long)b2 << 16) |
				((long)b3 << 8) | (long)b4);

	/* fi_mtime */
	b1 = (unsigned char)(fi.fi_mtime & 0x000000ff);
	b2 = (unsigned char)((fi.fi_mtime & 0x0000ff00) >> 8);
	b3 = (unsigned char)((fi.fi_mtime & 0x00ff0000) >> 16);
	b4 = (unsigned char)((fi.fi_mtime & 0xff000000) >> 24);
	fi.fi_mtime = (long)(((long)b1 << 24) | ((long)b2 << 16) |
				((long)b3 << 8) | (long)b4);

	/* fi_ctime */
	b1 = (unsigned char)(fi.fi_ctime & 0x000000ff);
	b2 = (unsigned char)((fi.fi_ctime & 0x0000ff00) >> 8);
	b3 = (unsigned char)((fi.fi_ctime & 0x00ff0000) >> 16);
	b4 = (unsigned char)((fi.fi_ctime & 0xff000000) >> 24);
	fi.fi_ctime = (long)(((long)b1 << 24) | ((long)b2 << 16) |
				((long)b3 << 8) | (long)b4);

	/* fi_count */
	b1 = (unsigned char)(fi.fi_count & 0x000000ff);
	b2 = (unsigned char)((fi.fi_count & 0x0000ff00) >> 8);
	b3 = (unsigned char)((fi.fi_count & 0x00ff0000) >> 16);
	b4 = (unsigned char)((fi.fi_count & 0xff000000) >> 24);
	fi.fi_count = (long)(((long)b1 << 24) | ((long)b2 << 16) |
				((long)b3 << 8) | (long)b4);
}
#endif	/* __INTEL */
